from django.shortcuts import render, get_object_or_404
from django.http import HttpResponse, Http404, HttpResponseRedirect
from django.template import loader
from django.urls import reverse, reverse_lazy
from django.views import generic, View
from django.forms import formset_factory, inlineformset_factory
from django.contrib.auth import logout, mixins, get_user
from django.contrib.auth.views import LoginView
from django.core.mail import send_mail
from . import models
from .forms import QuestionForm, OneChoiceForm
import datetime
            
class IndexView(generic.View):
    latest_question_list = models.Question.objects.order_by('-pub_date')[:5]
    cat_list = models.Category.objects.all()
    template_name = 'polls/index.html'
    
    def get(self, req, *args, **kwargs):
        context = {
                'latest_question_list':self.latest_question_list,
                'category_list':self.cat_list
                }
        return render(req, self.template_name, context)

def vote(request, question_id):
    print('choice called')
    question = get_object_or_404(models.Question, pk=question_id)
    print(question)
    try:
        selected_choice = question.choice_set.get(pk=request.POST['choice'])
        print('selected choice found')
    except (KeyError, models.Choice.DoesNotExist):
        # Redisplay the question voting form.
        return render(request, 'polls/detail.html', {
            'question': question,
            'error_message': "You didn't select a choice.",
        })
    else:
        selected_choice.votes += 1
        selected_choice.save()
        # Always return an HttpResponseRedirect after successfully dealing
        # with POST data. This prevents data from being posted twice if a
        # user hits the Back button.
        return HttpResponseRedirect(reverse('polls:results', args=(question.id,)))
    return HttpResponse("You're voting on question %s." % question_id)
    
class QuestionDetailView(generic.DetailView):
    model = models.Question
    template_name = 'polls/detail.html'

class QuestionResultsView(generic.DetailView):
    model = models.Question
    template_name = 'polls/results.html'

class AddCategoryView(mixins.LoginRequiredMixin, generic.edit.CreateView):
    # la CreateView va chercher un template_name = 'polls/category_form.html'
    model = models.Category
    fields = ['name']
    success_url = reverse_lazy('polls:index')

class CategoryView(generic.DetailView):
    model = models.Category
    question_model = models.Question
    template_name = 'polls/category.html'
    
    def get(self, req, *args, **kwargs):
        self.category_id = kwargs['category_id']
        category = get_object_or_404(self.model, pk=self.category_id)
        questions = self.question_model.objects.filter(category_id__id=self.category_id)
        return render(req, self.template_name, {'category':category , 'questions':questions})
    
    def get_context_data(self, **kwargs):
        context['category'] = model 
        return context
    
    def get_queryset(self, category_id):
        return self.model.filter(pk = category_id)

class ListCategoriesView(generic.ListView):
    temlate_name = 'polls/index.html'
    context_object_name = 'latest_quest'
    model = models.Category

    def get_queryset(self, category_id):
        return Question.objects.filter(category_id__id = category_id)

class AddQuestionViewOld(mixins.LoginRequiredMixin, mixins.PermissionRequiredMixin, View):
    """ une espece de form view faite à la main. On la garde pour des raisons pédagogique.
    la bonne mlanière de faire, c'est d'heriter de formview voire create view """

    permission_required = 'polls.can_think'
    form_class = QuestionForm
    template_name = 'polls/questionform.html'
    QuestionFormSet = formset_factory(OneChoiceForm, extra=2)

    def get(self, req, *args, **kwargs):
        form = self.form_class()
        context = {'form': form, 'formset':self.QuestionFormSet}
        return render(req, self.template_name, context)

    def post(self, req, *args, **kargs):
        print('### post ###')
        form = self.form_class(req.POST)
        formset = self.QuestionFormSet(req.POST, req.FILES) 
        if form.is_valid() and formset.is_valid():
            question = models.Question(
                question_text = form.cleaned_data['question_text'],
                pub_date = datetime.datetime.now(),
                category_id = form.cleaned_data['category']
                )
            question.save()
            user = self.request.user
            # maintenant on peut sauver les choix
            for form_choice in formset:
                print('cleaned data:{}'.format(form_choice.cleaned_data))
                if form_choice.cleaned_data:
                    choice = models.Choice(
                        question=question,
                        choice_text=form_choice.cleaned_data['choice'])
                    choice.save()
            send_mail(
                'Nouvelle question',
                'Une nouvelle question a été créée par {} : {}'.format(user, question.question_text),
                'from@example.com',
                ['antoine@kidivid.org'],
                fail_silently=False,
            )
            return HttpResponseRedirect(reverse(
                'polls:valid', 
                kwargs={'question_id':question.pk})
                )
        return render(req, self.template_name, {'form': forms})

class QuestionMixin:
    model = models.Question
    fields = ['question_text', 'category_id']
    success_url = reverse_lazy('polls:index')
    ChoiceFormSet = inlineformset_factory(
        models.Question, 
        models.Choice,
        fields=('choice_text',),
        can_delete=False,
        )
    
    def form_valid(self, form):
        if self.formset.is_valid() and form.is_valid():
            self.object = form.save()
            self.formset.instance = self.object
            self.formset.save()
        return super().form_valid(form)
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data()
        context['formset'] = self.formset
        return context

    def get_form(self, *args, **kwargs):
        self.formset = self.ChoiceFormSet(self.request.POST or None, instance=self.object)
        return super().get_form(*args, **kwargs)

class AddQuestionView(mixins.LoginRequiredMixin, mixins.PermissionRequiredMixin, QuestionMixin, generic.CreateView):
    template_name = 'polls/questionform.html'
    permission_required = 'polls.can_think'
    
class UpdateQuestionView(QuestionMixin, generic.UpdateView):
    template_name_suffix = '_update'
    
class DeleteQuestionView(generic.DeleteView):
    model = models.Question
    success_url = reverse_lazy('polls:index')

class PostRequestResult(View):
    template_name = 'polls/post_request_result.html'
    def get(self, req, *args, **kwargs):
        return render(req, self.template_name, {'question':kwargs['question_id']})

class AddChoiceView(mixins.LoginRequiredMixin, View):
    form_class = OneChoiceForm
    template_name = 'polls/add_choice.html'
    question = None

    def get(self, req, *args, **kwargs):
        self.question_id = kwargs['question_id']
        question = get_object_or_404(models.Question, pk=self.question_id)
        form = self.form_class()
        f1 = self.form_class.base_fields.get('first_choice')
        return render(req, self.template_name, {'form': form, 'question':question})

    def post(self, req, *args, **kwargs):
        form = self.form_class(req.POST)
        self.question_id = kwargs['question_id']
        print(self.question_id)
        question = get_object_or_404(models.Question, pk=self.question_id)
        if form.is_valid():
            choice = models.Choice(question=question, choice_text=req.POST['choice'])
            choice.save()
            
            return HttpResponseRedirect(reverse(
                'polls:add_choice', 
                kwargs={'question_id':self.question_id})
                )
        return render(req, self.template_name, {'form': form})

def logout_view(req):
    logout(req)
    return HttpResponseRedirect(reverse('polls:index'))

class PollLoginView(LoginView):
    def post(self, req, *args, **kwargs):
        super().post(req, *args, **kwargs)
        return HttpResponseRedirect(reverse('polls:index'))

class UsefullCategoriesView(generic.ListView): 
    template_name = 'polls/cat_list.html'
    
    def get_queryset(self):
        #choices = models.Choice.objects.filter(votes__gt = 2)
        # -------- ça marche, dans ce sens --------------------
        choices = models.Choice.objects.select_related(
            'question__category_id').filter(
                votes__gt = 2).distinct()
        # --- mais j'ai pas le résultat dans la queryset. il faut
        # --- il faut parcourir choices :
        for choice in choices:
            if choice.question.category_id:
                print(choice.question.category_id.name)

        # -------- on essaie en remontant dans l'autre sens plus dur ------
        # Category -> Question -> Choice

        # model_set, ça marche avec des objets (lignes de la table) :
        inst = models.Category.objects.get(pk=1)
        print(inst.question_set.all())

        # --- en fait la solution était simple avec filter : ---
        qs = models.Category.objects.filter(
                question__choice__votes__gte=2).distinct()

        return qs

class ChoiceDetail(View):
    template_name = 'polls/choice_detail.html'
    model = models.Choice
    
    def get(self, req, *args, **kwargs):
        self.choice_id = kwargs['pk']
        choice = self.model.objects.select_related('question__category_id').get(pk=kwargs['pk'])
        return render(req, self.template_name, {'choice':choice})
